#include <malloc.h>
#include <stdio.h>
#include "dict.h"

/* default dictionary settings: */
unsigned int alphabet_size = 27; /* 26 letters plus a space */
int (*index_function)(char *) = get_index;

dictnode *init_dict(unsigned int size, int (*indexer)(char *))
{
  if (size <= 0)
    {
      fprintf(stderr,"error: bad alphabet size: %d\n",size);
      exit(1);
    }
  if(size)                 /* this will have to change to support multiple */
    alphabet_size = size;  /* dictionaries with different sized nodes */

#ifdef DEBUG
fprintf(stderr,"init_dict: using size of %d\n",alphabet_size);
#endif

  if(indexer != NULL)        /* and different index functions */
    {
      index_function = indexer;
#ifdef DEBUG
      fprintf(stderr,"init_dict: indexer = %d\n",(unsigned int)indexer);
#endif
    }
#ifdef DEBUG
  else
    fprintf(stderr,"init_dict: using default indexer\n");
#endif

  return (newnode());    
}

dictnode *newnode()
{
  dictnode *node;
  int i;

  node = (dictnode *)malloc(sizeof(dictnode));
  if(node == NULL)
    {
      fprintf(stderr,"bad news!  could not malloc new dictionary node!\n");
      exit(1);
    }
  node->link =(struct dictstruct **)malloc(sizeof(dictnode *) * alphabet_size);
  node->word = 0;  /* no word ends here */
  node->id = -1;   /* an invalid id number */
  for(i=0; i<alphabet_size; i++) node->link[i] = NULL;
  
  return(node);
}

void dealloc_dict(dictnode *dict)
{
  int i;
  for(i=0; i < alphabet_size; i++)  /* deallocate this node's children, */
    if(dict->link[i] != (dictnode *)NULL)
      dealloc_dict(dict->link[i]);
  free(dict->link);                 /* the pointers to the ex-children, */
  free(dict);                       /* and then the node itself */
}

int get_index(char *c)  /* return number in 0..(alphabet_size - 1) */
{
  if(*c >= 'A' && *c <= 'Z') *c -= 32; /* lowercase *c */
  if(*c == ' ')
    return(alphabet_size - 1);  /* arbitrarily make spaces last */
  else if (*c >= 'a' && *c <= 'z')
    return(*c - 'a');     /* everything else is its pos. in alphabet */
  else
    return(-1);           /*  bad input character error code */
}

int insert_phrase(dictnode *dict, char *phrase, int id)
{
  char *c;
  dictnode *cursor;
  
  if (dict == NULL) return (-1);  /* can't insert into a null dictionary */
  
  cursor = dict;

  for(c = phrase; *c; c++)        /* better be null terminated strings   */
    {
      int index;

      index = (*index_function)(c);

      if(index < 0 || index >= alphabet_size)
	fprintf(stderr,"error! c = %c, index = %d, alphabet_size: %d\n",
		*c, index, alphabet_size);

      if (cursor->link[index] != NULL)  /* already is a link for this letter */
	{
	  cursor = cursor->link[index];  /* follow the link... */
	}
      else
	{
	  cursor->link[index] = newnode(); /* make a link for this letter */
	  cursor = cursor->link[index];    /* and follow it */
	}
    }
  cursor->word = 1;  /* finished the phrase, note that fact */
  cursor->id = id;   /* tag this node with its id number    */
  return(0);
}

int lookup(dictnode *dictionary, char *phrase)
{
  dictnode *cursor;    /* cursor into the dictionary */
  char *c;             /* an index into phrase */
  int index;           /* index into link[] */
  
  cursor = dictionary;
  c = phrase;
  while(*c)
    {
      index = (*index_function)(c);
      if (index == -1) return (-1); /* bad char in input */
      if (cursor->link[index] != NULL)  /* this letter matches */
	{
	  cursor = cursor->link[index]; /* follow the link */
	  c++;                          /* advance to next letter */
	}
      else  /* phrase doesn't match */
	{ 
	  return(-1);
	}
    }
  if(cursor->word == 1) return(cursor->id); else return(-1);
}

/* approximate lookup:  like lookup() but with error correcting heuristics 
 * 
 * the error correcting heuristics are really hairy.  but, that's the nature of
 * heuristics.  the algorithm follows the string down the dictionary.  so long
 * as everything matches, this is just like the lookup() function.  when a
 * letter fails to match the node pointed to by the cursor, the error
 * correcting gets invoked.  it starts by gathering some statistics about the
 * structure of the trie near the node the cursor points at with respect to the
 * next character in the string.
 *
 * specifically, it counts:
 * links_here:  the number of non-null links leading away from the node the
 *              cursor points at.  is something from 0 to alphabet_size.
 * links_match_next: the number of links at the node the cursor points at which
 *                   match the next letter.  is always 0 or 1.
 * links_next_match_this: if there is only one non-null link for the node the
 *                        cursor points at, then links_next_match_this is the
 *                        number of links at the next level down which match
 *                        the current letter.  also always 0 or 1.
 *
 * having gathered this information, the algorithm determines what kind of
 * error was committed.  only letter-transpositions, incorrect letters, and
 * missing errors can be dealt with.  all others errors are labeled as unknown.
 * also, some combinations of the above kinds of errors can result in an
 * unknown error.
 *
 * the error type is transpose if the next letter has a link at the node that
 * the cursor points to, and if that link leads to a node that matches the
 * current letter.  these determinations are made directly by inspecting the
 * trie, and don't involve the above statistics.
 *
 * if the error isn't a transpose, the algorithm refers to the following table
 * to assign an error type:

      links_match_next:
         |  0  |  1
       --+-----+----     ml = missing letter error
links  1 | ml/ | wl      wl = wrong letter error
here:    |  wl |
       --+-----+----
      >1 | ml  | wl

 * the table pairs values of links_here (rows) and links_match_next (columns).
 * error type determination is a simple lookup for (1,1), (>1,0), and (>1,1).
 * for (1,0), the error can be either a missing letter or a wrong letter.  to
 * resolve this, the algorithm checks the value of links_next_match_this. if
 * it is 1, the error is a missing letter.  if it is 0, the error is a wrong
 * letter.
 *
 * having determined the error type, the algorithm executes code appropriate
 * to fix that kind of error and get the regular string matching back on
 * track.
 *
 * note, the method of determining what type of error has occured isn't at
 * all intuitive. it was derived from a sample trie constructed from the words:
 *     abalone, able, food, fr bread, fr fries, gorilla, goryness
 * and various kinds of errors applied to those words during lookup.  if you
 * really want to understand why the values of links_here, etc, map onto error
 * types the way i've assigned them, you should construct the trie for those
 * seven words and try out mis-spellings of them yourself.
 *
 * as it turns out, there is a class of very bad mis-matches that occurs when
 * there are no branches in the local part of the trie and multiple wrong-
 * letters occur in a row, so there's also a check in the error-correcting
 * routine to catch this.
 */
int alookup(dictnode *dictionary, char *phrase, int *errorcount)
{
  dictnode *cursor;   /* cursor into the dictionary */
  char *c;            /* an index into phrase */
  int index;          /* index into link[] */
  int error_type = 0; /* 1 = missing letter, 2 = wrong letter, 3 = transpose */
  int error_on_last = 0; /* type of error on last past, if any */
  cursor = dictionary;
#ifdef DEBUG
  fprintf(stderr,"phrase: %s\n",phrase);
  fflush(stderr);
#endif
  if(cursor == NULL) { fprintf(stderr,"alookup: dictionary is NULL\n"); return -1; }
  c = phrase;
  while(*c)
    {
      index = (*index_function)(c);
      if (index == -1) return(-1);  /* bad char in input */
      if (cursor->link[index] != NULL)  /* this letter matches */
	{
	  cursor = cursor->link[index]; /* follow the link */
	  c++;                          /* advance to next letter */
          error_on_last = 0;            /* clear this for next pass */
	}
      else  /* phrase doesn't match; try to correct */
	{ 
	  int links_here = 0,       /* number of links at this node. */
	      links_match_next = 0, /* number of links at this node which go */
	                            /* to a match for the next letter. */
	      links_next_match_this = 0, /* is there a link at next level */
                                    /* that matches this letter, but only if */
                                    /* links_here == 1. */
	      one_link, one_link_n, /* which one link != NULL, for this node */
                                    /* and next node. note that these do not */
                                    /* hold useful or correct info if there */
                                    /* is more than one link at the node. */
	      one_link_next_match_this = 0, /* same, next node down matches */
                                            /* this letter */
	      n_index = (*index_function)(c+1), /* index of the next letter. */
	      i;                    /* a loop counter */
	  for(i = 0; i < alphabet_size; i++)  /* for each link in this node */
	    if(cursor->link[i] != NULL)     /* if the link points somewhere */
	      {
		links_here++;               /* count it */
		one_link = i;               /* remember which link is set */
		if (n_index != -1)          /* avoid bogus link check below */
		  {
#ifdef DEBUG
		    fprintf(stderr,"link[i]: %d i: %d n_index %d\n",
			    cursor->link[i],i,n_index);
		    fflush(stderr);
#endif
		    if((cursor->link[i])->link[n_index] != NULL)
		      {
			links_match_next++; /* next letter matches, count it */
			one_link_n = i;     /* remember which link is set */
		      }
		  }
		if((cursor->link[i])->link[index] != NULL)
		  {
		    links_next_match_this++;
		    one_link_next_match_this = i;
		  }
	      }

#ifdef DEBUG
fprintf(stderr,
	"c: %c, links_here: %d, links_match_next: %d, l_n_m_t: %d\n",
	*c,links_here,links_match_next,links_next_match_this);
#endif

	  if (n_index != -1 &&                        /* transpose error */
	      (((cursor->link[n_index] != NULL) &&
	      ((cursor->link[n_index])->link[index] != NULL))))
	    error_type = 3;

	  else if ( ( links_here >= 1 &&               /* wrong letter */
		      links_match_next == 1 ) ||
		    ( links_here == 1 &&
		      links_match_next == 0 &&
		      links_next_match_this == 0 ) )
	    error_type = 2;

	  else if ( links_here >=  1 &&               /* missing letter */
		    links_match_next == 0 &&
		    links_next_match_this == 1 )
	    error_type = 1;

	  else
	    error_type = 0;                            /* unknown error */
	  
	  switch(error_type)
	    {
	    case 0:
	      {
#ifdef DEBUG
 fprintf(stderr,"unknown error type\n");
#endif
		return(-1);
		break; }
	    case 1:
	      {
#ifdef DEBUG
 fprintf(stderr,"missing letter\n");
#endif
		if (links_here > 1 && links_next_match_this == 1)
		  cursor = cursor->link[one_link_next_match_this];
		else if (links_here == 1)
		  cursor = cursor->link[one_link]; 
		break; }
	    case 2:
	      {
#ifdef DEBUG
		fprintf(stderr,"wrong letter.\n");
#endif
		if(error_on_last == 2) /* there was a wrong letter last time */
                  return(-1);
		if (links_here > 1)
		  cursor = cursor->link[one_link_n];
		else if (links_here == 1)
		  cursor = cursor->link[one_link];
		c++;
		break; }
	    case 3:
	      {
#ifdef DEBUG
 fprintf(stderr,"transpose error.\n");
#endif
		cursor = (cursor->link[n_index])->link[index];
		c+=2;
		break; }
	    }
	  (*errorcount)++;            /* count the errors so far */
	  error_on_last = error_type; /* remember this for next time */
	}
    }

  if(cursor->word == 1)
    return(cursor->id);
  else
    return(-1);
}
